<?php

namespace Drupal\tmgmt_extension_suit\Form;

use Drupal\tmgmt\Form\JobForm;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt\JobItemInterface;

class JobMultiLangForm extends JobForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    unset($form['continuous_settings']);

    // This is a hidden form element that has ajax behaviour that updates
    // clone languages list. When user selects target language jquery
    // behaviour Drupal.behaviors.updateCloneLanguages will trigger 'click'
    // event for this element. Ajax callback of this element will return
    // updated version of 'update_clone_lang_options' fieldset with updated
    // languages list. See style.css and script.js. It's a textfield because
    // JobForm::validateForm() method expects
    // $form_state->getTriggeringElement()['#value'] to be string.
    $form['update_clone_lang_options'] = [
      '#type' => 'textfield',
      '#title' => t('Update clone languages list'),
      '#default_value' => 'update_clone_lang_options',
      '#ajax' => [
        'callback' => [$this, 'ajaxLanguagesUpdate'],
        'wrapper' => 'tmgmt-extension-suit-languages-wrapper',
        'event' => 'click',
      ],
    ];

    $form['clone_lang_options'] = [
      '#type' => 'details',
      '#title' => $this->t('Clone Job to these languages'),
      '#description' => $this->t('Select languages you would like to clone this Job to, during saving process.'),
      '#open' => TRUE,
      '#weight' => $form['clearfix']['#weight'] + 1,
      '#prefix' => '<div id="tmgmt-extension-suit-languages-wrapper">',
      '#suffix' => '</div>',
    ];

    $job = $this->entity;
    $source_language = $form_state->getValue('source_language') ?: $job->getSourceLangcode();
    $target_language = $form_state->getValue('target_language') ?: $job->getTargetLangcode();
    $langs = tmgmt_available_languages();

    unset($langs[$source_language]);
    unset($langs[$target_language]);

    $form['clone_lang_options']['languages'] = [
      '#title' => t('Languages'),
      '#type' => 'checkboxes',
      '#options' => $langs,
      '#required' => FALSE,
      '#prefix' =>  '<div class="check-controls"><span class="check-all">' . t('Check All') .'</span> ' .
        '<span class="uncheck-all">' . t('Uncheck All') .'</span></div>',
    ];

    // Attaching js libary for check-all/uncheck-all functionality.
    $form['#attached']['library'][] = 'tmgmt_extension_suit/check_uncheck_all';

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);

    $actions['save']['#submit'] = ['::submitForm', '::save'];

    if (isset($actions['submit'])) {
      $actions['submit']['#submit'] = ['::submitForm', '::save'];
    }

    return $actions;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    EntityForm::save($form, $form_state);

    $cloned_job_ids = $this->cloneJob($this->entity, array_filter($form_state->getValue('languages')));

    // Add cloned jobs to upload queue if tracking feature is turned on.
    foreach ($cloned_job_ids as $cloned_job_id) {
      \Drupal::service('tmgmt_extension_suit.utils.queue_unique_item')->addItem('tmgmt_extension_suit_upload', ['id' => (int) $cloned_job_id]);
    }

    // Everything below this line is only invoked if the 'Submit to provider'
    // button was clicked.
    if (isset($form['actions']['submit']) && $form_state->getTriggeringElement()['#value'] == $form['actions']['submit']['#value']) {

      // Delete conflicting items.
      if ($existing_items_ids = $form_state->get('existing_item_ids')) {
        $storage = \Drupal::entityTypeManager()->getStorage('tmgmt_job_item');
        $storage->delete($storage->loadMultiple($existing_items_ids));
        $num_of_items = count($existing_items_ids);
        drupal_set_message(\Drupal::translation()->formatPlural($num_of_items, '1 conflicting item has been dropped.', '@count conflicting items have been dropped.'), 'warning');
      }

      if (!tmgmt_job_request_translation($this->entity)) {
        // Don't redirect the user if the translation request failed but retain
        // existing destination parameters so we can redirect once the request
        // finished successfully.
        unset($_GET['destination']);
      }
      elseif ($redirect = tmgmt_redirect_queue_dequeue()) {
        // Proceed to the next redirect queue item, if there is one.
        $form_state->setRedirectUrl(Url::fromUri('base:' . $redirect));
      }
      elseif ($destination = tmgmt_redirect_queue_destination()) {
        // Proceed to the defined destination if there is one.
        $form_state->setRedirectUrl(Url::fromUri('base:' . $destination));
      }
      else {
        // Per default we want to redirect the user to the overview.
        $form_state->setRedirect('view.tmgmt_job_overview.page_1');
      }
    }
    else {
      // Per default we want to redirect the user to the overview.
      $form_state->setRedirect('view.tmgmt_job_overview.page_1');
    }
  }

  /**
   * Clones a given job to a given language.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   * @param $langs
   *
   * @return array
   */
  protected function cloneJob(JobInterface $job, $langs) {
    $cloned_job_ids = [];

    foreach ($langs as $lng) {
      if ($conflict_ids = $this->getConflictingItemIds($job, $job->getSourceLangcode(), $lng)) {
        $job->addMessage('This job wasn\'t cloned to "@lang" because of some conflicts with items: @ids', [
          '@lang' => $lng,
          '@ids' => implode(', ', $conflict_ids),
        ]);

        continue;
      }

      $new_job = $job->cloneAsUnprocessed();
      $new_job->uid = $this->currentUser()->id();
      $new_job->target_language = $lng;
      $new_job->save();

      // For now job file name copied from $job object.
      // We need to reset and regenerate.
      $new_job->set('job_file_name', '');

      // We should save job first and only then set state.
      // Job::preSave() relies on $this->original object.
      $new_job->setState(JobInterface::STATE_ACTIVE);

      /** @var \Drupal\tmgmt\JobItemInterface $item */
      foreach ($this->entity->getItems() as $item) {
        $item_to_resubmit = $item->cloneAsActive();
        $new_job->addExistingItem($item_to_resubmit);
      }

      $cloned_job_ids[] = $new_job->id();

      $job->addMessage('Job has been duplicated as a new job <a href=":url">#@id</a>.', [
        ':url' => $new_job->url(),
        '@id' => $new_job->id(),
      ]);

      $new_job->addMessage('This job is a duplicate of the job in another language <a href=":url">#@id</a>', [
        ':url' => $job->url(),
        '@id' => $job->id(),
      ]);
    }

    return $cloned_job_ids;
  }

  /**
   * Returns conflicting items.
   *
   * @param JobInterface $job
   *
   * @see \Drupal\tmgmt\Entity\Job::getConflictingItemIds for analog.
   *
   * @return array
   */
  protected function getConflictingItemIds(JobInterface $job, $src_lang, $target_lang) {
    $conflicting_item_ids = [];
    foreach ($job->getItems() as $item) {
      // Count existing job items that are have the same languages, same source,
      // are active or in review and are not the job item that we are checking.
      $existing_items_count = \Drupal::entityQuery('tmgmt_job_item')
        ->condition('state', [JobItemInterface::STATE_ACTIVE, JobItemInterface::STATE_REVIEW], 'IN')
        ->condition('plugin', $item->getPlugin())
        ->condition('item_type', $item->getItemType())
        ->condition('item_id', $item->getItemId())
        ->condition('tjiid', $item->id(), '<>')
        ->condition('tjid.entity.source_language', $src_lang)
        ->condition('tjid.entity.target_language', $target_lang)
        ->count()
        ->execute();

      // If there are any, this is a conflicting job item.
      if ($existing_items_count) {
        $conflicting_item_ids[] = $item->id();
      }
    }
    return $conflicting_item_ids;
  }

  /**
   * Ajax callback to update clone languages list.
   */
  public function ajaxLanguagesUpdate(array $form, FormStateInterface $form_state) {
    return $form['clone_lang_options'];
  }

}
