<?php

namespace Drupal\tmgmt_extension_suit\Tests;

use Drupal;
use Drupal\Core\Database\Database;
use Drupal\Core\Database\SchemaObjectExistsException;
use Drupal\Core\Queue\DatabaseQueue;
use Drupal\node\Entity\Node;
use Drupal\tmgmt\Entity\Job;
use Drupal\tmgmt\Tests\TMGMTTestBase;

/**
 * Basic tests for tmgmt extension module.
 *
 * @group tmgmt_extension_suit
 */
class TmgmtExtensionSuitTest extends TMGMTTestBase {

  /**
   * Logged in user.
   *
   * @var Drupal\user\UserInterface
   */
  private $userForTranslations;

  /**
   * Modules to enable.
   *
   * @var array
   */
  public static $modules = [
    'tmgmt',
    'tmgmt_demo',
    'tmgmt_extension_suit',
  ];

  /**
   * {@inheritdoc}
   */
  public function setUp() {
    parent::setUp();

    $this->userForTranslations = $this->loginAsAdmin([
      'edit any translatable_node content',
    ]);

    // Create queue table (it doesn't exist for simpletests somehow).
    $uploadQueue = new DatabaseQueue('tmgmt_extension_suit_upload', Database::getConnection());
    $database_schema = Drupal::database()->schema();

    try {
      if (!$database_schema->tableExists('queue')) {
        $schema_definition = $uploadQueue->schemaDefinition();
        $database_schema->createTable('queue', $schema_definition);
      }
    }
    catch (SchemaObjectExistsException $e) {
    }

    // Apply new column into tmgmt_job table.
    Drupal::service('entity.definition_update_manager')->applyUpdates();
  }

  /**
   * Requests nodes for translation.
   *
   * @param array $nids
   * @param $targetLanguage
   * @param $jobId
   */
  public function requestTranslation(array $nids, $targetLanguage, $jobId) {
    // Request nodes for translation.
    $data = [];

    foreach ($nids as $nid) {
      $data["items[$nid]"] = "item[$nid]";
    }

    $this->drupalPostForm('admin/tmgmt/sources', $data, t('Request translation'));

    $data = [
      'label[0][value]' => 'Test job',
      'target_language' => $targetLanguage,
      'translator' => 'local',
    ];
    $this->drupalPostForm("admin/tmgmt/jobs/$jobId", $data, t('Submit to provider'));
  }

  /**
   * Requests nodes for translation in a batch.
   *
   * @param array $nids
   * @param array $targetLanguages
   * @param $provider
   * @param array $settings
   * @throws \Exception
   */
  public function requestBulkTranslation(array $nids, array $targetLanguages, $provider = 'local', array $settings = []) {
    // Request nodes for translation.
    $data = [];

    foreach ($nids as $nid) {
      $data["items[$nid]"] = "item[$nid]";
    }

    $this->drupalPostForm('admin/tmgmt/sources', $data, t('Request translation in a batch'));

    if (!empty($targetLanguages)) {
      $new_data = [];

      foreach ($targetLanguages as $targetLanguage) {
        $new_data["target_language[$targetLanguage]"] = $targetLanguage;
      }

      $new_data['translator'] = $provider;

      foreach ($settings as $setting => $value) {
        $new_data[$setting] = $value;
      }

      $this->drupalPostForm(NULL, $new_data, t('Request translation'));
    }
  }

  /**
   * Translates job items (marks as "needs review").
   *
   * @param array $jobItemIds
   */
  public function translateJobItems(array $jobItemIds) {
    foreach ($jobItemIds as $jobItemId) {
      $edit = [
        'title|0|value[translation]' => $this->randomString(),
      ];
      $this->drupalPostForm("admin/tmgmt/items/$jobItemId", $edit, t('Save'));
    }
  }

  /**
   * Returns job item hash by node id and job id.
   *
   * @param $nid
   * @param $jobId
   *
   * @return bool|mixed
   */
  private function getNodeHash($nid, $jobId) {
    $hashQuery = Drupal::database()->select('tmgmt_job_item', 'tji');
    $hashQuery->join('tmgmt_job', 'tj', 'tj.tjid = tji.tjid');
    $hash = $hashQuery->condition('tj.tjid', $jobId)
      ->condition('tji.item_id', $nid)
      ->fields('tji', [
        'tjid',
        'tes_source_content_hash',
      ])
    ->execute()
    ->fetchAllKeyed();

    return $hash ? reset($hash) : FALSE;
  }

  /**
   * Returns amount of items in a given queue.
   *
   * @param string $queueName
   *
   * @return mixed
   */
  private function getCountOfItemsInQueue($queueName) {
    return Drupal::database()->select('queue', 'q')
      ->condition('q.name', $queueName)
      ->countQuery()
      ->execute()
      ->fetchField();
  }

  /**
   * Checks if job was added to a queue.
   *
   * @param $queue
   * @param $jobId
   *
   * @return mixed
   */
  private function isItemAddedToQueue($queue, $jobId) {
    return Drupal::database()->select('queue', 'q')
      ->condition('q.name', $queue)
      ->condition('q.data', "a:1:{s:2:\"id\";i:{$jobId};}")
      ->countQuery()
      ->execute()
      ->fetchField();
  }

  /**
   * Test "Track changes of the translatable entities" feature is turned on.
   */
  public function testTrackTranslatableEntityChanges() {
    $this->requestTranslation([1], 'fr', 1);

    // Submit node edit form without changes.
    // Expectations:
    // 1. Hash is not changed.
    // 2. Job is not added to upload queue.
    $oldHash = $this->getNodeHash(1, 1);
    $this->drupalPostForm("node/1/edit", [], t('Save'));
    $newHash = $this->getNodeHash(1, 1);
    $this->assertEqual($oldHash, $newHash);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 1), 0);

    // Submit node edit form with updated title.
    // Expectations:
    // 1. Hash is changed.
    // 2. Job is added to upload queue.
    $this->drupalPostForm("node/1/edit", [
      'title[0][value]' => 'New node test title',
    ], t('Save'));
    $newHash = $this->getNodeHash(1, 1);
    $this->assertNotEqual($oldHash, $newHash);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 1), 1);
  }

  /**
   * Test "Track changes of the translatable entities" feature is turned off.
   */
  public function testDoNotTrackTranslatableEntityChanges() {
    $this->requestTranslation([1], 'fr', 1);

    // Disable tracking feature.
    $this->drupalPostForm('admin/tmgmt/extension-settings', [
      'do_track_changes' => FALSE,
    ], t('Save configuration'));

    // Submit node edit form without changes.
    // Expectations:
    // 1. Hash is not changed.
    // 2. Job is not added to upload queue.
    $oldHash = $this->getNodeHash(1, 1);
    $this->drupalPostForm("node/1/edit", [], t('Save'));
    $newHash = $this->getNodeHash(1, 1);
    $this->assertEqual($oldHash, $newHash);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 1), 0);

    // Submit node edit form with updated title.
    // Expectations:
    // 1. Hash is not changed.
    // 2. Job is not added to upload queue.
    $this->drupalPostForm("node/1/edit", [
      'title[0][value]' => 'New node test title',
    ], t('Save'));
    $newHash = $this->getNodeHash(1, 1);
    $this->assertEqual($oldHash, $newHash);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 1), 0);
  }

  /**
   * Test tmgmt_extension_suit_add_to_check_status() function.
   *
   * In this test we don't run cron because it will immediately process
   * newly added queue items and we will not be able to check if added items
   * was correct. So we're calling a function that is called by hook_cron.
   */
  public function testCheckStatus() {
    $this->addLanguage('es');

    // Allow user to translate from en to es.
    $edit = [
      'tmgmt_translation_skills[0][language_from]' => 'en',
      'tmgmt_translation_skills[0][language_to]' => 'es',
    ];
    $this->drupalPostForm('user/' . $this->userForTranslations->id() . '/edit', $edit, t('Save'));

    // One job - one job item.
    $this->requestTranslation([1], 'fr', 1); // Job item 1.
    $this->requestTranslation([1], 'de', 2); // Job item 2.

    // One job - many job items.
    $this->requestTranslation([2, 3], 'fr', 3); // Job item 3, 4.
    $this->requestTranslation([2, 3], 'de', 4); // Job item 5, 6.
    $this->requestTranslation([2, 3], 'es', 5); // Job item 7, 8.

    // Let's mark job items as 'needs review'.
    $this->translateJobItems([2, 5, 7, 8]);

    // So now we have next jobs with job items:
    // job (state) - job item/s (state).
    // 1 (active) - 1 (active).
    // 2 (active) - 2 (review).
    // 3 (active) - 3 (active), 4 (active).
    // 4 (active) - 5 (review), 6 (active).
    // 5 (active) - 7 (review), 8 (review).
    // Expectation - in check status queue will be next jobs: 1, 3, 4
    tmgmt_extension_suit_add_to_check_status();

    $this->assertEqual($this->getCountOfItemsInQueue('tmgmt_extension_suit_check_status'), 3);

    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_check_status', 1), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_check_status', 2), 0);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_check_status', 3), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_check_status', 4), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_check_status', 5), 0);
  }

  /**
   * Test job cloning feature.
   */
  public function testJobCloning() {
    // Request translation for first node.
    $this->drupalPostForm('admin/tmgmt/sources', [
      'items[1]' => 'item[1]',
    ], t('Request translation'));

    // Translate to de and also clone to fr.
    $this->drupalPostForm('admin/tmgmt/jobs/1', [
      'target_language' => 'de',
      'languages[fr]' => 'fr',
    ], t('Submit to provider'));

    // In upload queue should be only one item.
    $this->assertEqual($this->getCountOfItemsInQueue('tmgmt_extension_suit_upload'), 1);

    // Cloned job with id = 2 should be in upload queue.
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 2), 1);
  }

  /**
   * Test validation for bulk translation.
   */
  function testRequestBulkTranslationValidation() {
    $settings = [
      'body' => [
        [
          'value' => $this->randomMachineName(32),
          'format' => filter_default_format(),
        ]
      ],
      'title' => $this->randomMachineName(8),
      'type' => 'translatable_node',
      'uid' => \Drupal::currentUser()->id(),
      'langcode' => 'fr',
    ];
    $node = Node::create($settings);
    $node->save();

    // Didn't select any nodes.
    $this->requestBulkTranslation([], []);
    $this->assertText("You didn't select any source items.");

    // Select nodes with different source languages.
    $this->requestBulkTranslation([3, $node->id()], []);
    $this->assertText("You can't request translation in a batch for items in different source languages.");

    // Select nodes with same source languages.
    $this->requestBulkTranslation([2, 3], []);
    $this->assertTitle("Request translation | Drupal");
  }

  /**
   * Test request bulk translation.
   */
  function testRequestBulkTranslation() {
    $this->requestBulkTranslation([2, 3], ['fr', 'de'], 'local', [
      'settings[translator]' => 1,
      'settings[job_comment]' => 'Comment',
    ]);

    $this->assertText('Translation request queued for 4 jobs. Return to "Sources" page.');

    // In upload queue should be 4 jobs: 2 nodes in 2 languages.
    $this->assertEqual($this->getCountOfItemsInQueue('tmgmt_extension_suit_upload'), 4);

    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 1), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 2), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 3), 1);
    $this->assertEqual($this->isItemAddedToQueue('tmgmt_extension_suit_upload', 4), 1);

    $jobs = Job::loadMultiple([1, 2, 3, 4]);

    // Check source language.
    $this->assertEqual($jobs[1]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[2]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[3]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[4]->getSourceLangcode(), 'en');

    // Check target language.
    $this->assertEqual($jobs[1]->getTargetLangcode(), 'de');
    $this->assertEqual($jobs[2]->getTargetLangcode(), 'fr');
    $this->assertEqual($jobs[3]->getTargetLangcode(), 'de');
    $this->assertEqual($jobs[4]->getTargetLangcode(), 'fr');

    // Check job settings.
    $this->assertEqual($jobs[1]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[2]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[3]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[4]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
  }

  /**
   * Test request bulk translation with conflicting item.
   */
  public function testRequestBulkTranslationSkippingNodeInOneLanguage() {
    // Let's get a conflicting job item.
    $this->requestTranslation([1], 'fr', 1);

    // Request bulk translation including conflicting job item.
    $this->requestBulkTranslation([1, 2, 3], ['fr'], 'local', [
      'settings[translator]' => 1,
      'settings[job_comment]' => 'Comment',
    ]);

    $this->assertText('1 item conflicted with pending item and was skipped. Existing job item:');
    $this->assertText('Job item with source language = "en" and target language = "fr" for entity "node" with id = "1".');
    $this->assertText('Translation request queued for 2 jobs. Return to "Sources" page.');

    // In upload queue should be 2 jobs: 2 nodes in fr.
    $this->assertEqual($this->getCountOfItemsInQueue('tmgmt_extension_suit_upload'), 2);

    $jobs = Job::loadMultiple([1, 2, 3]);

    // Check source language.
    $this->assertEqual($jobs[1]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[2]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[3]->getSourceLangcode(), 'en');

    // Check target language.
    $this->assertEqual($jobs[1]->getTargetLangcode(), 'fr');
    $this->assertEqual($jobs[2]->getTargetLangcode(), 'fr');
    $this->assertEqual($jobs[3]->getTargetLangcode(), 'fr');

    // Check job settings.
    $this->assertEqual($jobs[1]->get('settings')->getValue(), [
      [
        'translator' => '',
        'job_comment' => '',
      ],
    ]);
    $this->assertEqual($jobs[2]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[3]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
  }

  /**
   * Test request bulk translation with conflicting items.
   */
  public function testRequestBulkTranslationSkippingNodeInTwoLanguages() {
    // Let's get a conflicting job items.
    $this->requestTranslation([1], 'fr', 1);
    $this->requestTranslation([1], 'de', 2);

    // Request bulk translation including conflicting job items.
    $this->requestBulkTranslation([1, 2, 3], ['fr', 'de'], 'local', [
      'settings[translator]' => 1,
      'settings[job_comment]' => 'Comment',
    ]);

    $this->assertText('2 items conflicted with pending items and were skipped. Existing job items:');
    $this->assertText('Job item with source language = "en" and target language = "fr" for entity "node" with id = "1".');
    $this->assertText('Job item with source language = "en" and target language = "de" for entity "node" with id = "1".');
    $this->assertText('Translation request queued for 4 jobs. Return to "Sources" page.');

    // In upload queue should be 4 jobs: 2 nodes in fr and de.
    $this->assertEqual($this->getCountOfItemsInQueue('tmgmt_extension_suit_upload'), 4);

    $jobs = Job::loadMultiple([1, 2, 3, 4, 5, 6]);

    // Check source language.
    $this->assertEqual($jobs[1]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[2]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[3]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[4]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[5]->getSourceLangcode(), 'en');
    $this->assertEqual($jobs[6]->getSourceLangcode(), 'en');

    // Check target language.
    $this->assertEqual($jobs[1]->getTargetLangcode(), 'fr');
    $this->assertEqual($jobs[2]->getTargetLangcode(), 'de');
    $this->assertEqual($jobs[3]->getTargetLangcode(), 'de');
    $this->assertEqual($jobs[4]->getTargetLangcode(), 'fr');
    $this->assertEqual($jobs[5]->getTargetLangcode(), 'de');
    $this->assertEqual($jobs[6]->getTargetLangcode(), 'fr');

    // Check job settings.
    $this->assertEqual($jobs[1]->get('settings')->getValue(), [
      [
        'translator' => '',
        'job_comment' => '',
      ],
    ]);
    $this->assertEqual($jobs[2]->get('settings')->getValue(), [
      [
        'translator' => '',
        'job_comment' => '',
      ],
    ]);
    $this->assertEqual($jobs[3]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[4]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[5]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
    $this->assertEqual($jobs[6]->get('settings')->getValue(), [
      [
        'translator' => 1,
        'job_comment' => 'Comment',
      ],
    ]);
  }

}
